import 'dart:convert';

import 'crc.dart';
import 'models.dart';

class Payload {
  final String id;
  final String value;
  const Payload(this.id, this.value);

  int get length => value.length;

  @override
  String toString() {
    if (value.isEmpty) {
      return '';
    }
    return '$id${length.toString().padLeft(2, '0')}$value';
  }
}

/// Parse the [qrDataString] and orderly return id, value of the first data object, and the remainder string.
///
/// Return null if [qrDataString] is incompatible with the HKQR FPS specification.
///
/// Overview of structure of data within a HKQR:
///
/// Data within a QR code are organized in tree-like structure of data objects.
///
/// A data object may be a primitive data object or a template.
///
/// A template may include other primitive data objects and templates.
///
///
/// Each data object is made up of three individual fields.
///
/// The first field is an identifier (ID) by which the data object can be referenced.
///
/// The second field is a length field that explicitly indicates the number of characters
/// included in the third field, i.e. the value field.
///
/// A data object therefore comprises the following:
///
/// + ID field, which is coded as a two-digit numeric value, with a value ranging from "00" to "99";
/// + Length field, which is coded as a two-digit numeric value, with a value ranging from "01" to "99"; and
/// + Value field, which has a minimum length of one character and maximum length of 99 characters.
List<String>? parseFirstPartOfQrData(String qrDataString) {
  if (qrDataString.length < 5) return null;

  final String id = qrDataString.substring(0, 2);
  final int? valueLength = int.tryParse(qrDataString.substring(2, 4));

  if (valueLength == null) return null;

  if (qrDataString.length < valueLength + 4) return null;

  final String value = qrDataString.substring(4, 4 + valueLength);
  final String remainder = valueLength + 4 == qrDataString.length
      ? ''
      : qrDataString.substring(4 + valueLength, qrDataString.length);

  return [id, value, remainder];
}

class Validator {
  static final RegExp _alphanumericPattern = RegExp(r'^[a-zA-z0-9.-]+$');

  final int? exactLength;

  final int? upperLength;

  final bool alphanumeric;

  const Validator({this.exactLength, this.upperLength, this.alphanumeric = false});

  /// Validate if [content] has a valid length and is alphanumeric if expected.
  ///
  /// Return a non-null description if validate failed, or
  /// Return null if validate success.
  ///
  /// Null [content] will pass directly.
  String? validate(String content) {
    if (content.isEmpty) return null;
    final length = content.length;
    if (exactLength != null && length != exactLength) {
      return 'Length incorrect($exactLength)';
    }
    if (upperLength != null && length > upperLength!) {
      return 'Length exceeds limit($upperLength)';
    }
    if (alphanumeric && !_alphanumericPattern.hasMatch(content)) {
      return 'Alphanumeric value required';
    }
    return null;
  }
}

class TimeValidator extends Validator {
  const TimeValidator(int exactLength) : super(exactLength: exactLength);
  @override
  String? validate(String time) {
    final desc = super.validate(time);
    if (desc != null) return desc;

    /// yyyyMMddHHmmss -> 2012-02-27 13:27:00
    final year = '20${time.substring(0, 2)}';
    final month = time.substring(2, 4);
    final day = time.substring(4, 6);
    final hour = time.substring(6, 8);
    final min = time.substring(8, 10);
    final second = time.substring(10, 12);
    final timeOut = DateTime.tryParse('$year-$month-$day $hour:$min:$second');
    print('timeOut = $timeOut');
    if (timeOut == null) {
      return 'Merchant time-out is malformed';
    }
    if (timeOut.isBefore(DateTime.now())) {
      return 'Merchant time-out is ahead of current time';
    }
    return null;
  }
}

abstract class QrDataPart {
  QrDataPart._();

  List<Payload> encode();
  DecodeResult decode(String id, String template);
}

abstract class QrDataPartWithMandatoryIds implements QrDataPart {
  List<String> mandatoryIds = [];

  final _mandatoryPresence = <String, bool>{};

  QrDataPartWithMandatoryIds() {
    mandatoryIds.forEach((id) => _mandatoryPresence[id] = false);
  }

  DecodeResult successWithId(String id, {List<String> warnings = const []}) {
    _mandatoryPresence[id] = true;
    return DecodeResult.success(warnings: warnings);
  }

  bool allMandatoryPresent() => !_mandatoryPresence.values.contains(false);

  List<String> missedMandatoryIds() {
    List<String> missed = [];
    _mandatoryPresence.forEach((id, present) => present ? null : missed.add(id));
    return missed;
  }
}

class QRCodeConventions extends QrDataPartWithMandatoryIds {
  /// 00 Payload Format Indicator
  /// A fixed value of “01”
  static const String idPayloadFormat = '00';

  /// 01 Point of Initiation Method
  /// “11” for static QR Codes; “12” for dynamic QR Codes
  static const String idPointOfInitiation = '01';

  /// 63 Cyclic Redundancy Check
  /// Checksum calculated over all the data objects included in the QR code
  static const String idCrcCheck = '63';

  static const String prefix = '000201';

  static const _validators = {
    idPayloadFormat: Validator(exactLength: 2),
    idPointOfInitiation: Validator(exactLength: 2),
    idCrcCheck: Validator(exactLength: 4),
  };

  @override
  List<String> mandatoryIds = [idPayloadFormat];

  bool isDynamic;

  QRCodeConventions({this.isDynamic = false});

  List<Payload> encode() {
    return [
      Payload(idPayloadFormat, '01'),
      Payload(idPointOfInitiation, isDynamic ? '12' : '11'),
    ];
  }

  @override
  DecodeResult decode(String id, String template) {
    final desc = _validators[id]?.validate(template);
    if (desc != null) {
      return DecodeResult.failed(reason: desc);
    }
    switch (id) {
      case idPayloadFormat:
        if (template == '01') {
          return successWithId(id);
        } else {
          return DecodeResult.failed(
              reason: 'The Payload Format Indicator (ID “00”) '
                  'shall contain a value of "01". All other values are RFU.');
        }
      case idPointOfInitiation:
        switch (template) {
          case '12':
            isDynamic = true;
            return successWithId(id);
          case '11':
            isDynamic = false;
            return successWithId(id);
          default:
            return DecodeResult.failed(
                reason: 'The Point of Initiation Method (ID “01”) '
                    'shall contain a value of "11" or "12".');
        }
      case idCrcCheck:
        // Just swallow it. CRC check should be done somewhere else.
        return successWithId(id);
      default:
        return DecodeResult.unrecognized();
    }
  }

  @override
  bool allMandatoryPresent() {
    return true;
  }

  static String _crc(String input) {
    return Crc16CcittFalse().convert(Utf8Codec().encode(input)).toRadixString(16).padLeft(4, '0');
  }

  static String appendCrc(String input) {
    final withCrcId = '$input${idCrcCheck}04';
    final crc = _crc(withCrcId);
    return '$withCrcId$crc';
  }

  static bool crcCheck(String input) {
    if (input.length <= 8) return false;

    final contentToCrc = input.substring(0, input.length - 4);

    final crcId = input.substring(input.length - 8, input.length - 6);
    final crcLength = input.substring(input.length - 6, input.length - 4);
    final crcValue = input.substring(input.length - 4, input.length);

    if (crcId != idCrcCheck) return false;

    if (crcLength != '04') return false;

    return _crc(contentToCrc) == crcValue;
  }
}

class MerchantAccountInformation extends QrDataPartWithMandatoryIds {
  /// 26 Reserved for the Faster Payment System for use in Hong Kong
  ///
  /// from 02 to 51 except 26 are invalid on Payee-presented QR Code for fps
  static const String idMerchantAccountFps = '26';

  /// Globally Unique Identifier
  /// A fixed value of “hk.com.hkicl”, which is the reverse domain name of the operator of the FPS
  static const String idGloballyUid = '00';

  /// The Clearing Code of the Payee Participant.
  ///
  /// The Clearing Code (ID “01”) is optional if Phone Number or Email Address is used to
  /// identify the payee, depending on whether the payee uses the default FPS Participant
  /// or a designated FPS Participant to collect the payment.
  /// If FPS Identifier is used to identify the payee, the Clearing Code should not be specified.
  static const String idClearingCode = '01';

  /// FPS Identifier assigned by the FPS Addressing Service.
  /// Only one proxy identifier must be present.
  /// Maximum Length 34
  static const String idFpsIdentifier = '02';

  /// Phone number registered in the FPS Addressing Service.
  /// Only one proxy identifier must be present.
  /// Maximum Length 34
  static const String idPhoneNumber = '03';

  /// Email address registered in the FPS Addressing Service.
  /// Only one proxy identifier must be present.
  /// Maximum Length 34
  static const String idEmailAddress = '04';

  /// Specify in the format of “yymmddhhmmss”.
  ///
  /// For dynamic QR code (i.e. Point of Initiation Method = “12”),
  /// a merchant payee may specify the Merchant time-out time (ID “05”)
  /// to indicate the latest time which the merchant system will wait for the payment.
  static const String idMerchantTimeout = '05';

  /// Specify “1” when the Transaction Amount (ID “54”) is present and editable by payer customers.
  ///
  /// If the Transaction Amount (ID “54”) is present, the transaction amount by default is not
  /// editable by payer customers.
  /// The Editable Transaction Amount Indicator (ID “06”)
  /// can be used and the value of “1” should be specified if the transaction amount
  /// is editable by payer customers. The indicator should not be specified if the
  /// Transaction Amount (ID “54”) is not present.
  static const String idEditableTransactionAmountIndicator = '06';

  static const _validators = {
    idGloballyUid: Validator(exactLength: 12, alphanumeric: true),
    idClearingCode: Validator(exactLength: 3, alphanumeric: true),
    idFpsIdentifier: Validator(upperLength: 34, alphanumeric: true),
    idPhoneNumber: Validator(upperLength: 34, alphanumeric: true),
    idEmailAddress: Validator(upperLength: 34, alphanumeric: true),
    idMerchantTimeout: TimeValidator(12),
    idEditableTransactionAmountIndicator: Validator(exactLength: 1),
  };

  @override
  List<String> mandatoryIds = [idMerchantAccountFps, idGloballyUid];

  FpsProxy? fpsProxy;
  String timeout;
  String clearingCode;
  bool transactionAmountEditable;

  MerchantAccountInformation({
    this.fpsProxy,
    this.timeout = '',
    this.clearingCode = '',
    this.transactionAmountEditable = false,
  });

  Payload get fpsProxyPayload {
    if (fpsProxy == null) {
      return Payload(idPhoneNumber, '');
    }
    return Payload(
        fpsProxy!.isPhone
            ? idPhoneNumber
            : fpsProxy!.isEmail
                ? idEmailAddress
                : idFpsIdentifier,
        fpsProxy!.value);
  }

  List<Payload> encode() {
    return [
      Payload(
        idMerchantAccountFps,
        '${Payload(idGloballyUid, 'hk.com.hkicl')}'
        '$fpsProxyPayload'
        '${Payload(idClearingCode, clearingCode)}'
        '${Payload(idMerchantTimeout, timeout)}'
        '${Payload(idEditableTransactionAmountIndicator, transactionAmountEditable ? '1' : '')}',
      ),
    ];
  }

  @override
  DecodeResult decode(String id, String template) {
    if (id != idMerchantAccountFps) {
      final idInt = int.tryParse(id) ?? 0;

      /// Ids from '02' to '51' are valid payment operators id.
      /// But only FPS(ID 26) is supported.
      if (idInt >= 2 && idInt <= 51) {
        /// todo: One or more merchant account (IDs “02” to “51”) may be included
        return DecodeResult.failed(
          code: DecodeResult.errorUnsupportedPaymentOperator,
          reason: 'Unsupported payment operator: $id. \n Only FPS(ID 26) is supported.',
        );
      } else {
        return DecodeResult.unrecognized();
      }
    } else {
      List<String> warnings = [];
      String? phone, email, fpsId;
      String remainder = template;
      while (remainder.isNotEmpty) {
        final result = parseFirstPartOfQrData(remainder);
        if (result == null || result.length != 3) {
          return DecodeResult.failed(
            code: DecodeResult.errorFormatIncompatible,
            reason: 'Malformed data: $remainder',
          );
        }
        remainder = result[2];
        final innerId = result[0];
        final innerData = result[1];

        final desc = _validators[innerId]?.validate(innerData);
        if (desc != null) {
          return DecodeResult.failed(reason: '$desc: {ID: $id/$innerId, value: $innerData}');
        }
        switch (innerId) {
          case idGloballyUid:
            if (innerData != 'hk.com.hkicl') {
              return DecodeResult.failed(
                  reason: 'Globally Unique Identifier of fps must be hk.com.hkicl');
            }
            break;
          case idClearingCode:
            clearingCode = innerData;
            break;
          case idFpsIdentifier:
            fpsId = innerData;
            break;
          case idPhoneNumber:
            phone = innerData;
            break;
          case idEmailAddress:
            email = innerData;
            break;
          case idMerchantTimeout:
            timeout = innerData;
            break;
          case idEditableTransactionAmountIndicator:
            transactionAmountEditable = true;
            break;
          default:
            warnings.add('unknown MerchantAccountInformation id: id=$innerId, data=$innerData');
            break;
        }
        successWithId(innerId);
      }
      if (phone != null) {
        fpsProxy = FpsProxy(FpsType.phone, phone);
      }
      if (email != null) {
        if (fpsProxy == null) {
          fpsProxy = FpsProxy(FpsType.email, email);
        } else {
          return DecodeResult.failed(
              reason: 'Only one proxy identifier(phone, email or fpsId) should be specified');
        }
      }
      if (fpsId != null) {
        if (fpsProxy == null) {
          fpsProxy = FpsProxy(FpsType.fpsId, fpsId);
        } else {
          return DecodeResult.failed(
              reason: 'Only one proxy identifier(phone, email or fpsId) should be specified');
        }
      }
      if (fpsProxy == null) {
        return DecodeResult.failed(
            reason: 'One proxy identifier(phone, email or fpsId) must be present.');
      }
      if (clearingCode.isNotEmpty && fpsProxy!.isFpsId) {
        return DecodeResult.failed(
            code: DecodeResult.errorFormatIncompatible,
            reason: 'When FPS Identifier is used to identify the payee, '
                'the Clearing Code should not be specified');
      }
      return successWithId(id, warnings: warnings);
    }
  }
}

/// The Additional Merchant Information specifies the information about a merchant such as
/// merchant name and business location.
///
/// In general, a payee-presented QR code for FPS does not need to contain such information.
class AdditionalMerchantInformation extends QrDataPartWithMandatoryIds {
  /// Not used in payee-presented QR code for FPS.
  /// Populated with a dummy code of “0000” in this field.
  static const String idMerchantCategoryCode = '52';

  /// Not used in payee-presented QR code for FPS.
  /// Populated with a dummy code of “HK” or the actual country code in this field.
  static const String idCountryCode = '58';

  /// Not used in payee-presented QR code for FPS.
  /// Populated with a dummy code of “NA” in this field.
  static const String idMerchantName = '59';

  /// Not used in payee-presented QR code for FPS.
  /// Populated with a dummy code of “HK” or the actual city in this field.
  static const String idMerchantCity = '60';

  /// Not used in payee-presented QR code for FPS.
  static const String idPostalCode = '61';

  /// Not used in payee-presented QR code for FPS.
  static const String idLanguageTemplate = '64';

  static const _validators = {
    idMerchantCategoryCode: Validator(exactLength: 4),
    idCountryCode: Validator(exactLength: 2, alphanumeric: true),
    idMerchantName: Validator(upperLength: 25, alphanumeric: true),
    idMerchantCity: Validator(upperLength: 15, alphanumeric: true),
    idPostalCode: Validator(upperLength: 10, alphanumeric: true),
    idLanguageTemplate: Validator(exactLength: 99),
  };

  @override
  List<String> mandatoryIds = [
    idMerchantCategoryCode,
    idCountryCode,
    idMerchantName,
    idMerchantCity,
  ];

  @override
  List<Payload> encode() {
    return [
      Payload(idMerchantCategoryCode, '0000'),
      Payload(idCountryCode, 'HK'),
      Payload(idMerchantName, 'NA'),
      Payload(idMerchantCity, 'HK'),
    ];
  }

  @override
  DecodeResult decode(String id, String template) {
    final desc = _validators[id]?.validate(template);
    if (desc != null) {
      return DecodeResult.failed(reason: '$desc: {ID: $id, value: $template}');
    }
    switch (id) {
      case idMerchantCategoryCode:
        break;
      case idCountryCode:
        break;
      case idMerchantName:
        break;
      case idMerchantCity:
        break;
      case idPostalCode:
        break;
      case idLanguageTemplate:
        break;
      default:
        return DecodeResult.unrecognized();
    }
    return successWithId(id);
  }
}

class TransactionValue extends QrDataPartWithMandatoryIds {
  /// A numeric code based on [ISO 4217]. Only HKD (“344”) and CNY (“156”) are supported.
  static const String idTransactionCurrency = '53';

  /// e.g. “99.34”
  static const String idTransactionAmount = '54';

  /// Not used in payee-presented QR code for FPS.
  static const String idTipOrConvenienceIndicator = '55';

  /// Not used in payee-presented QR code for FPS.
  static const String idValueOfConvenienceFeeFixed = '56';

  /// Not used in payee-presented QR code for FPS.
  static const String idValueOfConvenienceFeePercentage = '57';

  static const _validators = {
    idTransactionCurrency: Validator(exactLength: 3),
    idTransactionAmount: Validator(upperLength: 13, alphanumeric: true),
    idTipOrConvenienceIndicator: Validator(exactLength: 2),
    idValueOfConvenienceFeeFixed: Validator(upperLength: 13, alphanumeric: true),
    idValueOfConvenienceFeePercentage: Validator(upperLength: 5, alphanumeric: true),
  };

  @override
  List<String> mandatoryIds = [idTransactionCurrency];

  bool isCurrencyHK;
  String amount;
  TransactionValue({this.isCurrencyHK = true, this.amount = ''});

  @override
  List<Payload> encode() {
    return [
      Payload(idTransactionCurrency, isCurrencyHK ? '344' : '156'),
      Payload(idTransactionAmount, amount),
    ];
  }

  @override
  DecodeResult decode(String id, String template) {
    final target = '{ID: $id, value: $template}';
    final desc = _validators[id]?.validate(template);
    if (desc != null) {
      return DecodeResult.failed(reason: '$desc: $target');
    }
    switch (id) {
      case idTransactionCurrency:
        if (template == '344') {
          isCurrencyHK = true;
        } else if (template == '156') {
          isCurrencyHK = false;
        } else {
          return DecodeResult.failed(reason: 'Only HKD(344) and CNY(156) are supported: $target');
        }
        break;
      case idTransactionAmount:
        if (double.tryParse(template) == null) {
          return DecodeResult.failed(reason: 'Amount is invalid: $target');
        }
        amount = template;
        break;
      case idTipOrConvenienceIndicator:
        break;
      case idValueOfConvenienceFeeFixed:
        break;
      case idValueOfConvenienceFeePercentage:
        break;
      default:
        return DecodeResult.unrecognized();
    }
    return successWithId(id);
  }
}

/// The Additional Data Template (ID “62”) is an optional template which includes
/// common additional data objects such as Bill Number and Reference Label of a merchant payment.
///
/// The EMV QRCPS standard also allows payment operators to define their own additional
/// data objects within this template, but it is not used in payee-presented QR code for FPS.
class AdditionalData implements QrDataPart {
  static const String rootId = '62';

  /// Either Bill Number or Reference Label can be put in the same payee-presented QR Code.
  static const String billNumber = '01';

  /// Either Bill Number or Reference Label can be put in the same payee-presented QR Code.
  static const String referenceLabel = '05';

  static const String mobileNumber = '02';
  static const String storeLabel = '03';
  static const String loyaltyNumber = '04';
  static const String customerLabel = '06';
  static const String terminalLabel = '07';
  static const String purposeOfTransaction = '08';
  static const String additionalConsumerDataRequest = '09';

  static const _validator = Validator(upperLength: 25, alphanumeric: true);
  static const _validators = {
    billNumber: _validator,
    referenceLabel: _validator,
    mobileNumber: _validator,
    storeLabel: _validator,
    loyaltyNumber: _validator,
    customerLabel: _validator,
    terminalLabel: _validator,
    purposeOfTransaction: _validator,
    additionalConsumerDataRequest: _validator,
  };

  String bill;
  String reference;

  AdditionalData({this.bill = '', this.reference = ''});

  @override
  List<Payload> encode() {
    return [
      Payload(
        rootId,
        '${Payload(billNumber, bill)}'
        '${Payload(referenceLabel, reference)}',
      )
    ];
  }

  @override
  DecodeResult decode(String id, String template) {
    if (id != rootId) {
      return DecodeResult.unrecognized();
    } else {
      List<String> warnings = [];
      String remainder = template;
      while (remainder.isNotEmpty) {
        final result = parseFirstPartOfQrData(remainder);
        if (result == null || result.length != 3) {
          return DecodeResult.failed(
            code: DecodeResult.errorFormatIncompatible,
            reason: 'Malformed data: $remainder',
          );
        }
        remainder = result[2];
        final innerId = result[0];
        final innerData = result[1];

        final desc = _validators[innerId]?.validate(innerData);
        if (desc != null) {
          return DecodeResult.failed(reason: '$desc: {ID: $id/$innerId, value: $innerData}');
        }

        switch (innerId) {
          case billNumber:
            bill = innerData;
            break;
          case referenceLabel:
            reference = innerData;
            break;
          default:

            /// all other ids are optional and not used in FPS
            break;
        }
      }
      return DecodeResult.success(warnings: warnings);
    }
  }
}
