import 'dart:convert' show ByteConversionSink, ByteConversionSinkBase, Converter;
import 'dart:typed_data';

class Crc16CcittFalse extends Converter<List<int>, CrcValue> {
  static List<int>? _table;
  final int _width = 16;
  final int _polynomial = 0x1021;
  final dynamic _initialValue = 0xffff;
  final dynamic _finalMask = 0x0000;

  Crc16CcittFalse() {
    if (_table == null) {
      _table = createByteLookupTable(_width, _polynomial);
    }
  }

  @override
  CrcValue convert(List<int> input) {
    final outputSink = FinalSink();
    final inputSink = startChunkedConversion(outputSink);
    inputSink.add(input);
    inputSink.close();
    return outputSink.value;
  }

  @override
  ByteConversionSink startChunkedConversion(Sink<CrcValue> outputSink) {
    return NormalSinkInt(
        _table as List<int>, _initialValue as int, _finalMask as int, outputSink, _width);
  }
}

/// Intermediate sink that performs the actual CRC calculation. It outputs to
/// [FinalSink].
abstract class _CrcSink<T> extends ByteConversionSinkBase {
  final List<T> table;
  final T finalMask;
  final Sink<CrcValue> _outputSink;
  late CrcLoopFunction _loopFunction;
  T value;
  int width;
  bool _closed;

  _CrcSink(this.table, this.value, this.finalMask, this._outputSink, this.width)
      : _closed = false,
        assert(value is int || value is BigInt) {
    _loopFunction = selectLoopFunction();
  }

  @override
  void add(List<int> chunk) {
    addSlice(chunk, 0, chunk.length, false /* isLast */);
  }

  @override
  void addSlice(List<int> chunk, int start, int end, bool isLast) {
    _loopFunction(chunk, start, end);
    if (isLast) {
      close();
    }
  }

  @override
  void close() {
    if (!_closed) {
      _closed = true;
      var v = (value as int) ^ (finalMask as int);
      _outputSink.add(CrcValue(width, v));
      _outputSink.close();
    }
  }

  CrcLoopFunction selectLoopFunction();
}

typedef CrcLoopFunction = void Function(List<int> chunk, int start, int end);

class CrcValue {
  final dynamic _value;
  final int _width;

  // BigInt values are ensured to be non-negative. But int values can go
  // negative due to the shifts and xors affecting the most-significant bit.
  CrcValue(this._width, this._value)
      : assert(_value is int || (_value is BigInt && !_value.isNegative));

  @override
  int get hashCode => _value.hashCode;

  @override
  bool operator ==(Object other) {
    if (other is CrcValue && _width == other._width) {
      return this == other._value;
    } else if (other is int) {
      if (_value is int) {
        return _value == other;
      }
      return BigInt.from(other).toUnsigned(_width) == _value;
    } else if (other is BigInt && !other.isNegative) {
      if (_value is BigInt) {
        return _value == other;
      }
      return BigInt.from(_value as int).toUnsigned(_width) == other;
    }
    return false;
  }

  @override
  String toString() {
    return toRadixString(10);
  }

  String toRadixString(int radix) {
    if (_value is int) {
      return (_value as int).toRadixString(radix);
    }
    return (_value as BigInt).toRadixString(radix);
  }
}

class NormalSinkInt extends NormalSink<int> {
  NormalSinkInt(List<int> table, int value, int finalMask, Sink<CrcValue> outputSink, int width)
      : super(table, value, finalMask, outputSink, width);

  void _crc8Loop(List<int> chunk, int start, int end) {
    for (final b in chunk.getRange(start, end)) {
      value = table[value ^ b];
    }
  }

  void _crc16Loop(List<int> chunk, int start, int end) {
    for (final b in chunk.getRange(start, end)) {
      value = table[(value >> 8) ^ b] ^ ((value << 8) & 0xFFFF);
    }
  }

  void _crc24Loop(List<int> chunk, int start, int end) {
    for (final b in chunk.getRange(start, end)) {
      value = table[(value >> 16) ^ b] ^ ((value << 8) & 0xFFFFFF);
    }
  }

  void _crc32Loop(List<int> chunk, int start, int end) {
    for (final b in chunk.getRange(start, end)) {
      value = table[(value >> 24) ^ b] ^ ((value << 8) & 0xFFFFFFFF);
    }
  }

  void _crcLoop(List<int> chunk, int start, int end) {
    final shiftWidth = width - 8;
    final mask = (1 << width) - 1;
    for (final b in chunk.getRange(start, end)) {
      value = table[((value >> shiftWidth) ^ b) & 0xFF] ^ ((value << 8) & mask);
    }
  }

  @override
  CrcLoopFunction selectLoopFunction() {
    switch (width) {
      case 32:
        return _crc32Loop;
      case 24:
        return _crc24Loop;
      case 16:
        return _crc16Loop;
      case 8:
        return _crc8Loop;
      default:
        return _crcLoop; // XXX: unused.
    }
  }
}

abstract class NormalSink<T> extends _CrcSink<T> {
  NormalSink(List<T> table, T value, T finalMask, Sink<CrcValue> outputSink, int width)
      : super(table, value, finalMask, outputSink, width);
}

class FinalSink extends Sink<CrcValue> {
  CrcValue? _value;

  CrcValue get value {
    assert(_value != null);
    return _value!;
  }

  @override
  void add(CrcValue data) {
    // Can only be called once.
    assert(_value == null);
    _value = data;
  }

  @override
  void close() {
    assert(_value != null);
  }
}

List<int> createByteLookupTable(int width, int poly) {
  assert(width >= 8);
  assert(poly is int);
  return _createByteLookupTableInt(width, poly);
}

List<int> _createByteLookupTableInt(int width, int poly) {
  var ret = Uint16List(256);
  final widthMask = (1 << width) - 1;
  final truncatedPoly = poly & widthMask;
  final topMask = 1 << (width - 1);
  for (var i = 0; i < 256; ++i) {
    int crc;
    crc = i << (width - 8);
    for (var j = 0; j < 8; ++j) {
      if ((crc & topMask) != 0) {
        crc = ((crc << 1) ^ truncatedPoly);
      } else {
        crc <<= 1;
      }
    }
    ret[i] = crc & widthMask;
  }
  return ret;
}
