import 'conventions.dart';
import 'models.dart';

/// Information represented by a FPS HKQR
class HKQrInfo {
  final bool isDynamic;

  /// merchant account info
  final FpsProxy? fpsProxy;
  final String timeout;
  final String clearingCode;
  final bool transactionAmountEditable;

  /// transaction value
  final bool isCurrencyHK;
  final String amount;

  /// additional data
  final String bill;
  final String reference;

  /// decode result, only valid when factory HKQrInfo.decode(String qrString)
  final DecodeResult decodeResult;

  final List<String> unrecognizedData;

  const HKQrInfo({
    this.isDynamic = false,
    this.fpsProxy,
    this.timeout = '',
    this.clearingCode = '',
    this.transactionAmountEditable = false,
    this.isCurrencyHK = true,
    this.amount = '',
    this.bill = '',
    this.reference = '',
    this.decodeResult = const DecodeResult.success(),
    this.unrecognizedData = const [],
  });

  factory HKQrInfo.decode(String qrString) {
    /// HKQR data must start with Payload Format Indicator
    if (!qrString.startsWith(QRCodeConventions.prefix)) {
      return HKQrInfo(
        decodeResult: DecodeResult.failed(
          code: DecodeResult.errorFormatIncompatible,
          reason: 'HKQR data must start with ${QRCodeConventions.prefix}',
        ),
      );
    }
    if (!QRCodeConventions.crcCheck(qrString)) {
      return HKQrInfo(
        decodeResult: DecodeResult.failed(
          code: DecodeResult.errorCrcCheckFailed,
          reason: 'CRC check failed',
        ),
      );
    }
    final qrCodeConventions = QRCodeConventions();
    final merchantAccountInformation = MerchantAccountInformation();
    final additionalMerchantInformation = AdditionalMerchantInformation();
    final transactionValue = TransactionValue();
    final additionalData = AdditionalData();
    final qrDataParts = [
      qrCodeConventions,
      merchantAccountInformation,
      additionalMerchantInformation,
      transactionValue,
      additionalData,
    ];
    final warnings = <String>[];
    final unRecognizedData = <String>[];
    String remainder = qrString;
    while (remainder.isNotEmpty) {
      final result = parseFirstPartOfQrData(remainder);
      if (result == null || result.length != 3) {
        return HKQrInfo(
          decodeResult: DecodeResult.failed(
            code: DecodeResult.errorFormatIncompatible,
            reason: 'Malformed data: $remainder',
          ),
        );
      }
      remainder = result[2];
      final id = result[0];
      final template = result[1];
      bool recognized = false;
      for (QrDataPart part in qrDataParts) {
        final result = part.decode(id, template);
        if (result.isUnrecognized) {
          continue;
        }
        recognized = true;
        if (result.isFailed) {
          return HKQrInfo(decodeResult: result);
        }
        if (result.isSuccess) {
          warnings.addAll(result.warnings);
        }
      }
      if (!recognized) {
        unRecognizedData.add('{id: $id, template: $template}');
      }
    }

    /// first part missed mandatory or null
    final partMissedMandatory = qrDataParts.fold<QrDataPartWithMandatoryIds?>(
      null,
      (p, e) => p == null && e is QrDataPartWithMandatoryIds && !e.allMandatoryPresent() ? e : p,
    );

    if (partMissedMandatory != null) {
      return HKQrInfo(
        decodeResult: DecodeResult.failed(
          code: DecodeResult.errorMandatoryMissed,
          reason: '$partMissedMandatory: mandatory ids missed: '
              '${partMissedMandatory.missedMandatoryIds()}',
        ),
      );
    }
    return HKQrInfo(
      isDynamic: qrCodeConventions.isDynamic,
      fpsProxy: merchantAccountInformation.fpsProxy,
      timeout: merchantAccountInformation.timeout,
      clearingCode: merchantAccountInformation.clearingCode,
      transactionAmountEditable: merchantAccountInformation.transactionAmountEditable,
      isCurrencyHK: transactionValue.isCurrencyHK,
      amount: transactionValue.amount,
      bill: additionalData.bill,
      reference: additionalData.reference,
      decodeResult: DecodeResult.success(warnings: warnings),
      unrecognizedData: unRecognizedData,
    );
  }

  String encode() {
    final qrDataParts = <QrDataPart>[
      QRCodeConventions(isDynamic: isDynamic),
      MerchantAccountInformation(
        fpsProxy: fpsProxy,
        timeout: timeout,
        clearingCode: clearingCode,
        transactionAmountEditable: transactionAmountEditable,
      ),
      AdditionalMerchantInformation(),
      TransactionValue(isCurrencyHK: isCurrencyHK, amount: amount),
      AdditionalData(bill: bill, reference: reference),
    ];
    final qrData =
        qrDataParts.fold<String>('', (p, e) => '$p${e.encode().fold('', (p, e) => '$p$e')}');
    return QRCodeConventions.appendCrc(qrData);
  }

  @override
  String toString() {
    if (decodeResult.isFailed) {
      return 'decodeFailed:\n'
          'code: ${decodeResult.resultCode} - ${decodeResult.codeDescription}\n'
          'reason: ${decodeResult.failedReason},\n';
    }
    return 'dynamic: $isDynamic \n\n'
        'fpsProxy: $fpsProxy,\n\n'
        'timeout: $timeout,\n\n'
        'editable: $transactionAmountEditable,\n\n'
        'currency: ${isCurrencyHK ? 'HKD' : 'RMB'},\n\n'
        'amount: $amount,\n\n'
        'clearingCode: $clearingCode,\n\n'
        'bill: $bill,\n\n'
        'reference: $reference,\n\n'
        'warnings: ${decodeResult.warnings},\n\n'
        'unrecognized: $unrecognizedData\n\n';
  }
}
