import 'dart:convert';

import 'package:flutter_test/flutter_test.dart';
import 'package:hkqr_fps/hkqr_fps.dart';
import 'package:hkqr_fps/src/conventions.dart';
import 'package:hkqr_fps/src/crc.dart';

void main() {
  crcTest();
  codecTest();
}

void crcTest() {
  String crc(String input) => Crc16CcittFalse()
      .convert(Utf8Codec().encode(input))
      .toRadixString(16)
      .padLeft(4, '0')
      .toUpperCase();

  test('crc', () {
    expect(crc('input'), '4A93');
    expect(crc('abcdacad'), 'D0EA');
    expect(crc('.'), '245C');
    expect(crc('a.b,c_d'), '9AA5');
    expect(crc(','), '041E');
  });
}

void codecTest() {
  final fpsProxy = FpsProxy(FpsType.phone, '12345678');
  final hkQrInfo = HKQrInfo(
    isDynamic: true,
    fpsProxy: fpsProxy,
    timeout: '211201010101',
    clearingCode: 'abc',
    transactionAmountEditable: true,
    isCurrencyHK: true,
    amount: '99.99',
    bill: 'bill_number',
    reference: 'reference',
  );

  HKQrInfo hkQrInfoDecoded;

  String payloadFormat = '000201';
  String pointOfInitialMethod = '010212';
  String merchantInfo = '0012hk.com.hkicl0308123456780103abc051221120101010106011';
  String additionalMerchantInfo = '520400005802HK5902NA6002HK';
  String currency = '5303344';
  String amount = '540599.99';
  String additionalData = '0111bill_number0509reference';

  test('encoder', () {
    expect(
      hkQrInfo.encode(),
      '$payloadFormat'
      '$pointOfInitialMethod'
      '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
      '$additionalMerchantInfo'
      '$currency$amount'
      '62${additionalData.length.toString().padLeft(2, '0')}$additionalData'
      '6304e635',
    );
  });

  test('decoder', () {
    String crced;
    final qrStringMissPointOfInitial = '$payloadFormat'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringMissPointOfInitial);
    // print('crced = $crced');
    // print('decoded = ${HKQrInfo.decode(crced)}');
    expect(HKQrInfo.decode(crced).decodeResult.isSuccess, true);

    final qrStringMissMerchantInfo = '$payloadFormat'
        '$pointOfInitialMethod'
        // '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringMissMerchantInfo);
    expect(HKQrInfo.decode(crced).decodeResult.resultCode, DecodeResult.errorMandatoryMissed);

    ///            '0012hk.com.hkicl0308123456780103abc051221120101010106011';
    merchantInfo = '0013hk.com.hkicl0308123456780103abc051221120101010106011';
    final qrStringLengthIncorrect = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringLengthIncorrect);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.failedReason.startsWith('Length incorrect'), true);

    ///            '0012hk.com.hkicl0308123456780103abc051221120101010106011';
    merchantInfo = '0012cn.com.hkicl0308123456780103abc051221120101010106011';
    final qrStringValueIncorrect = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringValueIncorrect);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.resultCode, DecodeResult.errorUnsupportedValue);

    merchantInfo = '0308123456780103abc051221120101010106011';
    final qrStringMissedMandatory = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringMissedMandatory);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.resultCode, DecodeResult.errorMandatoryMissed);

    //              0012hk.com.hkicl0308123456780103abc051221120101010106011
    merchantInfo = '0012hk.com.hkicl0308123456780103你好啊051221120101010106011';
    final qrStringFormatIncorrect = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringFormatIncorrect);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.resultCode, DecodeResult.errorUnsupportedValue);

    //              0012hk.com.hkicl0308123456780103abc051221120101010106011
    merchantInfo = '0012hk.com.hkicl0308123456780103abc051221120a01010106011';
    final qrStringTimeOutIncorrect = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringTimeOutIncorrect);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.resultCode, DecodeResult.errorUnsupportedValue);

    //              0012hk.com.hkicl0308123456780103abc051221120101010106011
    merchantInfo = '0012hk.com.hkicl0308123456780103abc051219120101010106011';
    final qrStringTimeOutBeforeNow = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringTimeOutBeforeNow);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.resultCode, DecodeResult.errorUnsupportedValue);

    //              0012hk.com.hkicl0308123456780103abc051221120101010106011
    merchantInfo = '0012hk.com.hkicl0208123456780103abc051221120101010106011';
    final qrStringClearingCodeWithFpsID = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringClearingCodeWithFpsID);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.resultCode, DecodeResult.errorFormatIncompatible);

    /// restore to correct value
    merchantInfo = '0012hk.com.hkicl0308123456780103abc051221120101010106011';

    ///              '520400005802HK5902NA6002HK';
    additionalMerchantInfo = '5802HK5902NA6002HK';
    final qrStringMissedMandatory2 = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringMissedMandatory2);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.resultCode, DecodeResult.errorMandatoryMissed);

    ///              '520400005802HK5902NA6002HK';
    additionalMerchantInfo = '520400005802HK5902NA6016HK12345671234567';
    final qrStringLengthExceedsUpperLimit = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringLengthExceedsUpperLimit);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.failedReason.startsWith('Length exceeds limit'), true);

    /// restore to correct value
    additionalMerchantInfo = '520400005802HK5902NA6002HK';

    currency = '5303345';
    final qrStringCurrencyIncorrect = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringCurrencyIncorrect);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.resultCode, DecodeResult.errorUnsupportedValue);

    /// restore to correct value
    currency = '5303344';

    final qrStringMissedMandatory3 = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        //'$currency$amount'
        '$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringMissedMandatory3);
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    expect(hkQrInfoDecoded.decodeResult.resultCode, DecodeResult.errorMandatoryMissed);

    final qrStringCrcIncorrect = '$payloadFormat'
        '$pointOfInitialMethod'
        '26${merchantInfo.length.toString().padLeft(2, '0')}$merchantInfo'
        '$additionalMerchantInfo'
        '$currency$amount'
        '62${additionalData.length.toString().padLeft(2, '0')}$additionalData';
    crced = QRCodeConventions.appendCrc(qrStringCrcIncorrect);
    crced = crced.substring(0, crced.length - 1) + 'g';
    print('crced = $crced');
    hkQrInfoDecoded = HKQrInfo.decode(crced);
    print('hkQrInfoDecoded: $hkQrInfoDecoded');
    expect(hkQrInfoDecoded.decodeResult.resultCode, DecodeResult.errorCrcCheckFailed);
  });
}
